import tensorflow as tf
from tensorflow import keras
from tensorflow.keras.models import Sequential
from tensorflow.keras.layers import *
from tensorflow.keras.constraints import Constraint
import tensorflow.keras.backend as K
import numpy as np
import os
import pandas as pd
import matplotlib.pyplot as plt
from utils import *
tf.config.run_functions_eagerly(True)

size = 1152
sampling_layer_id = 1
original_model_path = os.path.abspath('original_CS_DL_model_path')
model_weigths_path = os.path.abspath('original_CS_DL_model_weigths_path')
refined_model_path = os.path.abspath('refined_CS_DL_model')
sampling_weights_path =  os.path.abspath('CS_DL_sampling_weigths.out')
refined_inference_net_path = os.path.abspath('CS_DL_refined_inference_subnet')

class DiagonalWeight(Constraint):
    def __call__(self, w):
        return w*(K.eye(K.int_shape(w)[-1]))

def define_model():
    model = Sequential()
    model.add(keras.layers.Flatten())
    model.add(keras.layers.Dense(1152, kernel_constraint=DiagonalWeight(), use_bias=False))
    model.add(keras.layers.Reshape(target_shape=(128, 9)))
    model.add(keras.layers.Conv1D(filters=64, kernel_size=3, activation='relu',  dtype='float32', padding='same', input_shape = (128,9)))
    model.add(keras.layers.BatchNormalization())
    model.add(keras.layers.Conv1D(filters=64, kernel_size=3, activation='relu',  dtype='float32', padding='same'))
    model.add(keras.layers.BatchNormalization())
    model.add(keras.layers.MaxPooling1D(pool_size=2))
    model.add(keras.layers.Conv1D(filters=64, kernel_size=3, activation='relu',  dtype='float32', padding='same'))
    model.add(keras.layers.BatchNormalization())
    model.add(keras.layers.MaxPooling1D(pool_size=2))
    model.add(keras.layers.Flatten())
    model.add(keras.layers.Dense(100, activation='relu'))
    model.add(keras.layers.BatchNormalization())
    model.add(keras.layers.Dense(6, activation='softmax'))
    return model

def define_inference_subnet():
    model = Sequential()
    model.add(keras.layers.Reshape(target_shape=(128, 9)))
    model.add(keras.layers.Conv1D(filters=64, kernel_size=3, activation='relu',  dtype='float32', padding='same', input_shape = (128,9)))
    model.add(keras.layers.BatchNormalization())
    model.add(keras.layers.Conv1D(filters=64, kernel_size=3, activation='relu',  dtype='float32', padding='same'))
    model.add(keras.layers.BatchNormalization())
    model.add(keras.layers.MaxPooling1D(pool_size=2))
    model.add(keras.layers.Conv1D(filters=64, kernel_size=3, activation='relu',  dtype='float32', padding='same'))
    model.add(keras.layers.BatchNormalization())
    model.add(keras.layers.MaxPooling1D(pool_size=2))
    model.add(keras.layers.Flatten())
    model.add(keras.layers.Dense(100, activation='relu'))
    model.add(keras.layers.BatchNormalization())
    model.add(keras.layers.Dense(6, activation='softmax'))
    return model

def define_sampling_subnet():
    model = Sequential()
    model.add(keras.layers.Flatten())
    model.add(keras.layers.Dense(1152, use_bias=False))
    return model

def make_predictions(model, trainX, trainy, batch_size):
    train_dataset = tf.data.Dataset.from_tensor_slices((trainX, trainy))
    train_dataset = train_dataset.shuffle(buffer_size=1024).batch(batch_size)
    prediction_size = trainX.shape[1]*trainX.shape[2]

    column_names = list(range(prediction_size))
    for i in range(len(column_names)):
        column_names[i] = str(i+1)

    column_names2 = list(range(trainy.shape[1]))
    for i in range(len(column_names2)):
        column_names2[i] = str(i+1)

    s = np.empty((0, prediction_size))
    l = np.empty((0, trainy.shape[1]))

    for i,(x, y) in enumerate(train_dataset):    
        samples = model(x, training=False)      
        p = np.append(s, np.array(samples), axis=0)
        r = np.append(l, np.array(y), axis=0)
        l = r
        s = p

    dataframe1 = pd.DataFrame(s, columns=column_names)
    dataframe2 = pd.DataFrame(l, columns=column_names2)

    dataframe1.to_csv('CS_samples_train.csv', mode='a', header=False, index = None)
    dataframe2.to_csv('CS_labels_train.csv', mode='a', header=False, index = None)

def train_original_CS_DL_model(trainX, trainy, valX, valy, testX, testy, batch_size, original_model_path):
    model = define_model()
    model = compile_model(model)
    
    EPOCHS = 100
    early_stopping_monitor = tf.keras.callbacks.EarlyStopping(
        monitor='val_loss',
        min_delta=0,
        patience=3,
        verbose=0,
        mode='min',
        baseline=None,
        restore_best_weights=True
    )

    checkpoint_filepath = '/tmp/checkpoint'
    model_checkpoint_callback = tf.keras.callbacks.ModelCheckpoint(
        filepath=checkpoint_filepath,
        save_weights_only=True,
        monitor='val_accuracy',
        mode='max',
        save_best_only=True)
    
    history = model.fit(
        x=trainX,
        y=trainy,
        validation_data=(valX, valy),
        epochs=EPOCHS,
        callbacks=[model_checkpoint_callback, early_stopping_monitor], 
    )
    metrics = history.history
    plt.plot(history.epoch, metrics['loss'], metrics['val_loss'])
    plt.legend(['loss', 'val_loss'])
    plt.show()
    plt.savefig("training_loss_CS_DL_model")

    test_acc = model.evaluate(testX, testy, batch_size=batch_size)
    print(test_acc)
    model.save_weights(original_model_path)
    w1 = model.layers[sampling_layer_id].get_weights()[0]
    np.savetxt(sampling_weights_path, w1, delimiter=' ')
    return model

def make_predictions_with_sampling_subnet(trainX, trainy, batch_size):
    sampling_subnet = define_sampling_subnet()
    sampling_subnet = compile_model(sampling_subnet)
    sampling_subnet(tf.keras.Input((128,9)))
    sampling_subnet.load_weights(original_model_path)
    sparsity_arr = [0, 0.1, 0.2, 0.3, 0.4, 0.5, 0.6, 0.7, 0.8, 0.9]
    for sparsity in sparsity_arr:
        model_pruned = set_group_weights_to_zero(sampling_subnet, sampling_layer_id, sparsity)
        make_predictions(model_pruned, trainX, trainy, batch_size)

def refine_inference_subnet():
    model = define_inference_subnet()
    dataframe1 = pd.read_csv('CS_samples_train.csv')
    dataframe1 = dataframe1.replace(np.nan, '', regex=True)
    dataframe1 = dataframe1[1:]
    x = tf.convert_to_tensor(dataframe1)
    dataframe2 = pd.read_csv('CS_labels_train.csv')
    dataframe2 = dataframe2.replace(np.nan, '', regex=True)
    dataframe2 = dataframe2[1:]
    y = tf.convert_to_tensor(dataframe2)

    trainX, testX, trainY, testY = train_test_split(np.array(x, dtype=np.float32), np.array(y, dtype=np.float32), test_size=0.2, shuffle=True)
    model = compile_model(model)
    callback = tf.keras.callbacks.EarlyStopping(monitor='loss', patience=3)
    history = model.fit(trainX, trainY, batch_size=64, shuffle=True, epochs=30, validation_split=0.2, callbacks=[callback])
    metrics = history.history
    plt.plot(history.epoch, metrics['loss'], metrics['val_loss'])
    plt.legend(['loss', 'val_loss'])
    plt.show()
    plt.savefig("training_loss_CS_DL_refined_model")

    model.save_weights(refined_inference_net_path)
    test_acc = model.evaluate(x=testX, y=testY, batch_size=64)
    print(test_acc)

def create_refined_model(testX, testy, batch_size):     
    new_model = define_model()
    new_model = compile_model(new_model)
   
    data = np.loadtxt(sampling_weights_path)
    w_shape = np.array(new_model.layers[1].get_weights()).shape

    data = np.reshape(data, w_shape)
    new_model.layers[1].set_weights(data)

    inference_subnet = define_inference_subnet()
    inference_subnet.load_weights(refined_inference_net_path)
    inference_subnet = compile_model(inference_subnet)
    inference_subnet(tf.keras.Input((1152)))
   
    i = 3
    for j in range(1,13):
        new_model.layers[i].set_weights(inference_subnet.layers[j].get_weights())
        i = i+1
    
    new_model.save_weights(refined_model_path)
    new_model.save(refined_model_path)

    new_model.summary()
    test_accuracy = new_model.evaluate(x=testX, y=testy, batch_size=batch_size)
    print(test_accuracy)
    return new_model
   
def load_model(testX, testy, batch_size): 
    model = define_model()
    model = compile_model(model) 
    model.load_weights(refined_model_path)
    model = compile_model(model)
    test_accuracy = model.evaluate(x=testX, y=testy, batch_size=batch_size)
    print(test_accuracy)
    model.summary()
    return model

def evaluate_different_sampling_rates(model, testX, testy, batch_size):
    for i in range(0, 10):
        perc = i/10
        model_pruned = set_group_weights_to_zero(model, sampling_layer_id, perc)
        model_pruned.evaluate(testX, testy, batch_size=batch_size)  
        filename = "pruned_model"+ str(i)
        model.save(filename)


def main():
    batch_size = 64
    trainX, trainy, valX, valy, testX, testy = load_dataset()
    #Step 1: train the original CS DL model
    CS_DL_model = train_original_CS_DL_model(trainX, trainy, valX, valy, testX, testy, batch_size, original_model_path)
    #Step 2: detach the sampling subnetwork and use it to generate inputs with various sparsity degrees coresponding to different sampling rates
    make_predictions_with_sampling_subnet(trainX, trainy, batch_size)
    #Step 3: re-fine the inference subnetwork with the inputs generated at step 2
    refine_inference_subnet()
    #Step 4: re-asamble the entire CS DL refined model
    refined_CS_DL_model = create_refined_model(testX, testy, batch_size)
    #refined_CS_DL_model = load_model(testX, testy, batch_size)
    #Step 5: test the refined CS DL model at various "sampling rates"
    evaluate_different_sampling_rates(refined_CS_DL_model, testX, testy, batch_size)

main()

