import numpy as np
from tensorflow import keras
from scipy.stats import entropy
from sklearn import svm
from sklearn.model_selection import train_test_split
from utils import *

def compute_entropy_thresholds_and_margins(model, testX, testy):
    data = []
    margins = []
    entropy_thresholds = []

    for i in range(0, 10):
        perc = i/10
        model_pruned = set_group_weights_to_zero(model, layer_id=1, percentage=perc)
        
        y_pred_model_w_temp = model_pruned.predict(testX)
    
        e = []
        l = []
        e_corr = []
        e_incorr = []
        j = 0
        n = np.log(np.power(2,6))

        l = np.zeros((testX.shape[0]))
        for arr in y_pred_model_w_temp:
            if (np.max(np.array(arr)) and (np.argmax(np.array(arr)) == testy[j].argmax())):
                e_corr.append(entropy(np.array(arr))/n)
                l[j] = 0
            else:
                e_incorr.append(entropy(np.array(arr))/n)
                l[j] = 1
            j = j+1
            e.append(entropy(np.array(arr))/n)
        data.append(e_corr)
   
        grid = np.zeros((np.array(e).shape[0],2))
        for j in range(0, np.array(e).shape[0]):
            grid[j,0] = e[j]
            grid[j,1] = 0

        e = np.array(e)
        x = e.reshape((e.shape[0], 1))
        y = np.zeros((e.shape[0], 1))
        grid = np.hstack((x, y))

        grid1 = np.zeros((np.array(e_corr).shape[0],2))
        for j in range(0, np.array(e_corr).shape[0]):
            grid1[j,0] = e_corr[j]
            grid1[j,1] = 0

        grid2 = np.zeros((np.array(e_incorr).shape[0],2))
        for j in range(0, np.array(e_incorr).shape[0]):
            grid2[j,0] = e_incorr[j]
            grid2[j,1] = 0

        X_train, X_test, y_train, y_test = train_test_split(grid, l, test_size=0.2,  shuffle=True, random_state=True)
        clf = svm.SVC(kernel='linear', class_weight='balanced')

        X_train = np.array(X_train)
        X_test = np.array(X_test)

        clf = clf.fit(X_train, y_train)
        w = clf.coef_[0]
        x_0 = -clf.intercept_[0]/w[0]
        margin = 1/w[0]
        x_min, x_max = np.floor(X_train.min()), np.ceil(X_train.max())
        y_min, y_max = -1, 1
        yy = np.linspace(y_min, y_max)

        XX, YY = np.mgrid[x_min:x_max:200j, y_min:y_max:200j]
        Z = clf.predict(np.c_[XX.ravel(), np.zeros(XX.size)]).reshape(XX.shape)
        # plt.clf()
        # plt.figure(dpi=200, figsize=(25, 7), tight_layout=True)
        # plt.style.use('default')
        # plt.plot(x_0*np.ones(shape=yy.shape), yy, 'k-')
        # plt.plot(x_0*np.ones(shape=yy.shape) - 0.8*margin, yy, 'k--')
        # plt.plot(x_0*np.ones(shape=yy.shape) + 0.8*margin, yy, 'k--')
        # plt.scatter(grid2[:, 0], grid2[:, 1], facecolors='none', c='orange', s=500, marker='^', label="Incorrect predictions")
        # plt.scatter(grid1[:, 0], grid1[:, 1], facecolors='none', c='steelblue', s=500,  marker='^', label="Correct predictions")
        # plt.xlim(x_min, x_max)
        # plt.ylim(y_min, y_max)
        # plt.yticks([])
        # plt.tick_params(axis='both', which='major', labelsize=42)
        # plt.legend(prop={"size":42})
        # plt.savefig("E_SVM_boundary_x"+str(i))
        margins.append(margin)
        entropy_thresholds.append(x_0)
    return entropy_thresholds, margins

def run_adaptation(list_models, testX, testy, entropy_thresholds_models, margins):
    acc = np.zeros((10))
    samples = np.zeros((10))

    mha = [2,2,5,5,9,7] #for each of the 6 activity classes, the best accuracy model
    mhc = [6,7,8,9,9,9] #for each of the 6 activity classes, the best compression model with acceptable accuracy)

    model_index = 9
    current_model = list_models[model_index]

    for i in range(0, testX.shape[0]):  
        x = np.array(testX[i,:])
        x = x.reshape(1, x.shape[0]*x.shape[1])
       
        y_pred = current_model(np.expand_dims(testX[i,:,:],0), training=False)
       
        e = entropy(np.array(y_pred).flatten())/1.7917

        y_pred = np.argmax(np.array(y_pred))

        if (y_pred == np.array(testy[i,:]).argmax()):
            acc[model_index] += 1
        samples[model_index] += 1

        if (e <= entropy_thresholds_models[model_index] + margins[model_index]):
            model_index = mhc[y_pred]
        else:
            model_index = mha[y_pred]
        current_model = list_models[model_index]      
    
    avg_model = 0
    for i in range(0, len(list_models)):
        print("model ", i, acc[i], samples[i])
        avg_model += i*samples[i]
    print(avg_model/testX.shape[0])
    print(np.sum(acc)/np.sum(samples))

trainX, trainy, valX, valy, testX, testy = load_dataset()
CS_DL_model = keras.models.load_model("pruned_model0")
entropy_thresholds_models, margins = compute_entropy_thresholds_and_margins(CS_DL_model, testX, testy)

model0 = keras.models.load_model("pruned_model0")
model1 = keras.models.load_model("pruned_model1")
model2 = keras.models.load_model("pruned_model2")
model3 = keras.models.load_model("pruned_model3")
model4 = keras.models.load_model("pruned_model4")
model5 = keras.models.load_model("pruned_model5")
model6 = keras.models.load_model("pruned_model6")
model7 = keras.models.load_model("pruned_model7")
model8 = keras.models.load_model("pruned_model8")
model9 = keras.models.load_model("pruned_model9")
   
list_models = [] 
list_models.append( model0 )
list_models.append( model1 )
list_models.append( model2 )
list_models.append( model3 )
list_models.append( model4 )
list_models.append( model5 )
list_models.append( model6 )
list_models.append( model7 )
list_models.append( model8 )
list_models.append( model9 )

run_adaptation(list_models, testX, testy, entropy_thresholds_models, margins)